from enum import IntEnum
from os import PathLike
from typing import List, Self


class WatchEventType(IntEnum):  # undocumented
    """
    A type of watch event.
    """
    
    Create = 0
    Delete = 1
    Modify = 2


class WatchEvent:  # undocumented
    """
    A watch event.
    """
    
    kind: WatchEventType
    target: str


class Watch:  # undocumented
    """
    Watch a directory and its subdirectory for changes.
    """
    
    def __init__(self): ...
    def add_watch(self, to_watch: PathLike | str, is_recursive: bool): ...
    def __aiter__(self) -> Self: ...
    async def __anext__(self) -> List[WatchEvent]: ...


import asyncio
import logging
from json import dumps
from pathlib import Path
from typing import Annotated
from typing import TypeAlias

import typer

from wakapi_anyide import __version__
from wakapi_anyide.models.config import WakatimeConfig
from wakapi_anyide.models.environment import Environment
from wakapi_anyide.models.project import Project
from wakapi_anyide.runner import ConfigInvalidatedException
from wakapi_anyide.runner import run

logger = logging.getLogger(__name__)


DEFAULT_IGNOREFILES = [".gitignore"]
TEMPLATE = """
# https://github.com/iamawatermelo/wakapi-anyide v{version}

[meta]
version = 1
watchers = ['files']

[files]
include = {include}  # files to include in tracking
exclude = {exclude}  # files to exclude in tracking
exclude_files = {exclude_files}  # files whose contents will be used to exclude other files from tracking
exclude_binary_files = true  # whether to ignore binary files
# language_mapping = {{".kicad_sch" = "Kicad Schematic"}} # custom language mapping

[project]
name = "{name}"  # your project name
"""

app = typer.Typer(
    pretty_exceptions_enable=False  # they don't report asyncio taskgroup exceptions correctly
)

try:
    from rich.highlighter import Highlighter
    from rich.text import Text
    
    class NoHighlights(Highlighter):
        def highlight(self, text: Text) -> None:
            pass
except ImportError:
    pass


def setup_logging(is_verbose: bool):
    try:
        from rich.logging import RichHandler
        logging.basicConfig(
            level="DEBUG" if is_verbose else "INFO",
            format="[bold magenta]{module}[/bold magenta][bright_black]:{funcName}@{lineno:03}[/bright_black]  {message}",
            style='{',
            datefmt="[%X]",
            handlers=[RichHandler(highlighter=NoHighlights(), rich_tracebacks=True, markup=True, show_time=False, show_path=False)]
        )
    except ImportError:
        logging.basicConfig(
            level="DEBUG" if is_verbose else "INFO",
            format="{levelname} {filename}:{funcName}@{lineno:03} {message}",
            style='{',
            datefmt="[%X]"
        )
        logger.warning("Rich is not available, using basic logging (pip install rich?)")


Verbose: TypeAlias = Annotated[bool, typer.Option("--verbose", callback=setup_logging)]


def start(is_test):
    asyncio.run(run(Environment(
        is_test_only=is_test,
        config=WakatimeConfig(),  # type: ignore
        project=Project()  # type: ignore
    )))


@app.command()
def test(verbose: Verbose = False):
    start(True)


@app.command()
def track(verbose: Verbose = False):
    start(False)


@app.command()
def version():
    print(f"wakapi-anyide v{__version__}")

def prompt(prompt, default: str | None = None):
    try:
        from rich import get_console
        return get_console().input(f"[bold magenta]{prompt}[/bold magenta] [cyan](default {default})[/cyan]\n[bright_cyan]>>> [bright_cyan]") or default
    except ImportError:
        return input(f"{prompt}\n>>> ") or default


def prompt_choices(prompt, choices, default):
    try:
        from rich import get_console
        response = get_console().input(f"[bold magenta]{prompt} {repr(choices)}[/bold magenta] [cyan](default {default})[/cyan]\n[bright_cyan]>>> [bright_cyan]")
    except ImportError:
        response = input(f"{prompt} {repr(choices)} (default {default})\n>>> ")
    
    if response == "":
        return default
    
    if response in choices:
        return response
    else:
        raise ValueError("response not allowed")


def prompt_yn(prompt, default: bool):
    prompt_str = "[Y/n]" if default else "[y/N]"
    
    try:
        from rich import get_console
        response = get_console().input(f"[bold magenta]{prompt}[/bold magenta] [cyan]\\{prompt_str}: [cyan]")
    except ImportError:
        response = input(f"{prompt} {prompt_str}: ")
        
    if response == "":
        return default
    
    if response.lower() not in ("y", "n"):
        raise ValueError("expected y or n")
    
    return response.lower() == "y"


@app.command()
def setup():
    output = Path("wak.toml")
    if output.exists():
        raise Exception("a wak.toml already exists in this directory")
        
    project_type = prompt_choices("What kind of project do you have?", ["files"], "files")
        
    project_name = prompt("What's your project name?", default=Path("./").absolute().name)
    
    included_paths = list()
    if prompt_yn("Would you like to watch all files in the directory?", True):
        included_paths.append("*")
    elif prompt_yn("Would you like to add include paths?", True):
        while True:
            included_paths.append(prompt("Please enter a path to include in gitignore format (e.g /src)"))
            
            if not prompt_yn("Would you like to add another include path?", True):
                break
    
    excluded_paths = list()
    if prompt_yn("Would you like to add exclude paths?", False):
        while True:
            excluded_paths.append(prompt("Please enter a path to exclude in gitignore format (e.g /node_modules)"))
            
            if not prompt_yn("Would you like to add another exclude path?", False):
                break
    
    exclude_files = []
    for file in DEFAULT_IGNOREFILES:
        if Path(file).exists():
            exclude_files.append(file)
    
    with open(output, 'w') as file:
        file.write(TEMPLATE.format(
            version=__version__,
            include=dumps(included_paths),
            exclude=dumps(excluded_paths),
            exclude_files=dumps(exclude_files),
            name=project_name
        ).strip())


if __name__ == "__main__":
    app()



import asyncio
import base64
import logging
import time
from asyncio import CancelledError
from asyncio import Future
from asyncio import Queue
from asyncio import Task
from asyncio import TaskGroup
from collections.abc import Sequence
from hashlib import sha256
from platform import uname
from typing import Dict

from aiohttp import ClientResponse
from aiohttp import request

from wakapi_anyide.models.environment import Environment
from wakapi_anyide.watchers import WATCHERS
from wakapi_anyide.watchers.types import Event
from wakapi_anyide.watchers.types import Watcher

logger = logging.getLogger(__name__)


class ConfigInvalidatedException(Exception):
    pass
    

async def heartbeat_task(env: Environment, queue: Queue[Event], watchers: Sequence[Watcher], should_shutdown: asyncio.Event):
    next_heartbeat_due = time.time() + env.config.settings.heartbeat_rate_limit_seconds
    fut: Future[Event] | None = None

    while not should_shutdown.is_set():
        changed_events: Dict[str, Event] = dict()
        
        while (due := next_heartbeat_due - time.time()) > 0 and not should_shutdown.is_set():
            if fut is None or fut.done():
                fut = asyncio.create_task(queue.get())
            
            logger.debug(f"Next event due in {due}s")
            
            completed, rest = await asyncio.wait([
                fut,
                asyncio.create_task(should_shutdown.wait())
            ], return_when=asyncio.FIRST_COMPLETED, timeout=due)  # type: ignore
            
            if fut in completed:
                event = fut.result()

                logger.debug(f"Got event for {event.filename}!")

                changed_events[event.filename] = event
        
        logger.debug("Processing heartbeats")
        
        next_heartbeat_due = time.time() + env.config.settings.heartbeat_rate_limit_seconds
        
        for watcher in watchers:
            logger.debug(f"Getting events from {watcher}")
            iterable = watcher.resolve_events()
            logger.debug(f"Maybe iterable is {iterable}")
            
            if iterable is not None:           
                async for event in iterable:
                    changed_events[event.filename] = event
                    logger.debug(f"Got event for {event.filename}")
        
        logger.debug(changed_events)
        
        if len(changed_events) == 0:
            logger.debug(f"No changes detected.")
            continue

        logger.info(f"Change summary:")
        for event in changed_events.values():
            logger.info(f"{event.filename:20} at {event.cursor[0]}:{event.cursor[1]} +{event.lines_added} -{event.lines_removed}")

        host = uname()
        user_agent = f"wakatime/unset ({host.system}-none-none) wakapi-anyide-wakatime/unset"

        heartbeats = [{
            "entity": event.filename,
            "type": "file",
            "category": "coding",
            "time": event.time,
            "project": env.project.project.name,
            "language": language_processor(env, event.file_extension),
            "lines": event.lines,
            "line_additions": event.lines_added,
            "line_deletions": event.lines_removed,
            "lineno": event.cursor[0],
            "cursorpos": event.cursor[1],
            "is_write": True,
            "editor": "wakapi-anyide",
            "machine": env.config.settings.hostname or f"anonymised machine {sha256(host.node.encode()).hexdigest()[:8]}",
            "operating_system": host.system,
            "user_agent": user_agent
        } for event in changed_events.values()]

        response: ClientResponse
        last_text: str | None = None

        if env.is_test_only:
            logger.info(f"Would've sent heartbeat, but in testing mode")
            continue

        for i in range(3):
            logger.info(f"Sending heartbeat (attempt {i+1})")
            async with request("POST", f"{env.config.settings.api_url}/users/current/heartbeats.bulk", json=heartbeats, headers={
                "User-Agent": user_agent,
                "Authorization": f"Basic {base64.b64encode(env.config.settings.api_key.encode()).decode()}"
            }) as response:
                if response.status == 201:
                    break
                else:
                    last_text = await response.text()
        else:
            raise Exception(f"Failed to send heartbeat: {response.status} {last_text}")


def language_processor(env: Environment, file_extension: str) -> str:
    languages = env.project.files.language_mapping

    lang = languages.get(file_extension)  # If the suffix matches a defined one in the languages table
    if lang is None:
        return file_extension.replace(".", "")  # If it didn't find a match, return the suffix only
    return lang
    
    
async def run(env: Environment):
    ev = asyncio.get_event_loop()
    runners = [WATCHERS[watcher](env) for watcher in env.project.meta.watchers]
    emit_events: Queue[Event] = Queue()
    should_shutdown = asyncio.Event()
    task: Future | None = None
    
    try:
        async with TaskGroup() as tg:
            for runner in runners:
                await runner.setup(tg, emit_events)
            
            async def throw_exception(exc):
                raise exc
            
            task = ev.create_task(heartbeat_task(env, emit_events, runners, should_shutdown))
            
            def done_callback(task: Task):
                logger.debug(task)
                try:
                    exc = task.exception()
                except CancelledError:
                    return
                
                if exc is not None and type(exc) not in (CancelledError, KeyboardInterrupt):
                    print(exc)
                    tg.create_task(throw_exception(exc))
        
            task.add_done_callback(done_callback)
    except KeyboardInterrupt:
        pass
    finally:
        should_shutdown.set()
        if task is not None:
            await task



import configparser
from pathlib import Path
from typing import Any
from typing import Dict

from pydantic_settings import BaseSettings
from pydantic_settings.sources import ConfigFileSourceMixin
from pydantic_settings.sources import InitSettingsSource


class IniConfigSettingsSource(InitSettingsSource, ConfigFileSourceMixin):
    _config_dict: Dict[str, Any]
    
    def __init__(self, settings_cls: type[BaseSettings]):
        file = settings_cls.model_config.get("ini_file")
        assert file
        
        super().__init__(settings_cls, self._read_files(file))
    
    def _read_file(self, path: Path) -> dict[str, Any]:
        parsed = configparser.ConfigParser()
        parsed.read(path)
        return {section: dict(parsed[section]) for section in parsed.sections()}



import difflib
import logging
from dataclasses import dataclass
from pathlib import Path

from aiofiles import open
from aiofiles.ospath import getsize

from wakapi_anyide.models.environment import Environment
from wakapi_anyide.watchers.types import Event

logger = logging.getLogger(__name__)

SIZE_MAX = 2**16


@dataclass
class File:
    path: str
    body: bytes
    size: int
    
    @property
    def too_large(self):
        return self.size > SIZE_MAX
    
    @classmethod
    async def read(cls, path: str):
        size = await getsize(path)
        
        if size > SIZE_MAX:
            return cls(
                path,
                b"",
                size
            )
        else:
            async with open(path, 'rb') as file:
                return cls(
                    path,
                    await file.read(),
                    size
                )
    
    @classmethod
    def empty(cls, path: str):
        return cls(
            path,
            b"",
            0
        )


def index_to_linecol(file: str, index: int):
    line = 0
    col = 0

    for character in file[:index]:
        if character == '\n':
            line += 1
            col = 0
        else:
            col += 1

    return line + 1, col + 1


def process_file_change(new_file: File, old_file: File, time: float, env: Environment) -> Event | None:
    filename = new_file.path
    file_extension = Path(new_file.path).suffix
    
    if new_file.too_large or old_file.too_large:
        diff = new_file.size - old_file.size
        lines_added = max(0, diff)
        lines_removed = -min(0, diff)
        
        return Event(
            filename=f"{filename}#wakapi-anyide-toolarge",
            file_extension=file_extension,
            cursor=(0, 0),
            lines_added=lines_added,
            lines_removed=lines_removed,
            lines=new_file.size,
            time=time
        )
    
    try:
        new_file_str = new_file.body.decode()
        old_file_str = old_file.body.decode()

        last_index = 0
        for op in difflib.SequenceMatcher(a=old_file_str, b=new_file_str, autojunk=False).get_opcodes():
            match op:
                case ('replace', _, _, _, j2):
                    last_index = max(last_index, j2)
                case ('delete', i1, _, _, _):
                    last_index = max(last_index, i1)
                case ('insert', i1, _, j1, j2):
                    last_index = max(last_index, j2)
                case ('equal', _, _, _, _):
                    pass
                case _:
                    raise Exception(f"Unknown opcode {op}")

        new_file_lines = new_file_str.splitlines()
        added_lines = 0
        deleted_lines = 0
        for op in difflib.SequenceMatcher(a=old_file_str.splitlines(), b=new_file_lines, autojunk=False).get_opcodes():
            match op:
                case ('replace', i1, i2, j1, j2):
                    added_lines += j2 - j1
                    deleted_lines += i2 - i1
                case ('delete', i1, i2, _, _):
                    deleted_lines += i2 - i1
                case ('insert', _, _, j1, j2):
                    added_lines += j2 - j1
                case ('equal', _, _, _, _):
                    pass
                case _:
                    raise Exception(f"Unknown opcode {op}")

        line, col = index_to_linecol(new_file_str, last_index)

        return Event(
            filename=filename,
            file_extension=file_extension,
            cursor=(line, col),
            lines_added=added_lines,
            lines_removed=deleted_lines,
            lines=len(new_file_lines),
            time=time
        )
    except UnicodeDecodeError:
        if env.project.files.exclude_binary_files:
            logger.info(f"Ignored file {filename}")
            return

        added_lines = 0
        deleted_lines = 0
        last_index = 0
        for op in difflib.SequenceMatcher(a=old_file.body, b=new_file.body, autojunk=False).get_opcodes():
            match op:
                case ('replace', i1, i2, j1, j2):
                    added_lines += j2 - j1
                    deleted_lines += i2 - i1
                    last_index = max(last_index, j2)
                case ('delete', i1, i2, _, _):
                    deleted_lines += i2 - i1
                    last_index = max(last_index, i1)
                case ('insert', _, _, j1, j2):
                    added_lines += j2 - j1
                    last_index = max(last_index, j2)
                case ('equal', _, _, _, _):
                    pass
                case _:
                    raise Exception(f"Unknown opcode {op}")

        return Event(
            filename=filename,
            file_extension=f"{filename}#wakapi-anyide-binaryfile", # custom handling for binary files
            cursor=(1, last_index),
            lines_added=added_lines,
            lines_removed=deleted_lines,
            lines=len(new_file.body),
            time=time
        )



from typing import List
from typing import Literal

from pydantic import BaseModel
from pydantic_settings import BaseSettings
from pydantic_settings import TomlConfigSettingsSource
from pydantic_settings.main import SettingsConfigDict


class ProjectMeta(BaseModel):
    version: Literal[1]
    watchers: List[str] = ['files']


class ProjectFiles(BaseModel):
    include: List[str]
    exclude: List[str]
    exclude_files: List[str]
    exclude_binary_files: bool = True
    language_mapping: dict = {}


class ProjectDescription(BaseModel):
    name: str


class Project(BaseSettings):
    model_config = SettingsConfigDict(toml_file="wak.toml")
    
    meta: ProjectMeta
    files: ProjectFiles
    project: ProjectDescription
    
    @classmethod
    def settings_customise_sources(cls, settings_cls, *_args, **_kwargs):
        return (TomlConfigSettingsSource(settings_cls), )



from dataclasses import dataclass

from wakapi_anyide.models.config import WakatimeConfig
from wakapi_anyide.models.project import Project


@dataclass
class Environment:
    is_test_only: bool
    config: WakatimeConfig
    project: Project



import logging
import time
from asyncio import Task
from asyncio.queues import Queue
from asyncio.taskgroups import TaskGroup
from collections.abc import AsyncGenerator
from pathlib import Path
from typing import Dict

from aiofiles import open
from pathspec import PathSpec

from wakapi_anyide._rust.watch import Watch
from wakapi_anyide._rust.watch import WatchEventType
from wakapi_anyide.helpers.filediffer import File
from wakapi_anyide.helpers.filediffer import process_file_change
from wakapi_anyide.models.environment import Environment
from wakapi_anyide.watchers.types import Event
from wakapi_anyide.watchers.types import Watcher

logger = logging.getLogger()

def bytes_to_human(size: int):
    if size == 0:
        return "(empty)"
    
    for i, suffix in enumerate(["B", "KiB", "MiB", "GiB", "TiB"]):
        if 1024**i < size < 1024**(i+1):
            return f"{(size / (1024**i)):.2f} {suffix}"
    
    return f"{(size / 1024**6):.2f} PiB"
    

def format_file(file: File):
    color = "red" if file.too_large else "bright_black"
    extra = "  [yellow]Large file[/yellow]" if file.too_large else ""
    return(f"{file.path}  [{color}]{bytes_to_human(file.size)}[/{color}]{extra}")


class FileWatcher(Watcher):
    """
    Watches filesystem events for file changes.
    """
    
    current_file: File | None = None
    env: Environment
    task: Task
    cache: Dict[str, File] = dict()
    
    def __init__(self, env: Environment):
        self.env = env
        
    @staticmethod
    def normalise(path: Path):
        return f"./{(path.relative_to(Path('./').absolute()))}"
        
    async def _task(self, queue: Queue[Event]):
        excluded_pathspecs = self.env.project.files.exclude.copy()
    
        for file in self.env.project.files.exclude_files:
            async with open(file, 'r') as file:
                excluded_pathspecs.extend(await file.readlines())
    
        included_paths = PathSpec.from_lines('gitwildmatch', self.env.project.files.include)
        excluded_paths = PathSpec.from_lines('gitwildmatch', excluded_pathspecs)
    
        watch = Watch()
        watch.add_watch("./", True)
        
        logger.info("Watched files:")
        
        for path in included_paths.match_tree('./'):
            resolved_path = self.normalise(Path('./').absolute() / Path(path))
            
            if excluded_paths.match_file(resolved_path):
                continue
            
            try:
                file = await File.read(resolved_path)
            except Exception as e:
                logger.error(e)
                continue
            
            self.cache[resolved_path] = file
            logger.info(f"  {format_file(file)}")
    
        logger.info("Watching!")
    
        async for ev_list in watch:
            for event in ev_list:
                logger.debug(f"Got event: {event}")
                resolved_path = self.normalise(Path(event.target))
    
                if not included_paths.match_file(resolved_path) or excluded_paths.match_file(resolved_path):
                    continue
                
                if event.kind == WatchEventType.Delete:
                    new_file = File.empty(resolved_path)
                else:
                    try:
                        new_file = await File.read(resolved_path)
                    except OSError as e:
                        if not Path(resolved_path).is_dir():
                            logger.warning(f"Failed to open a file: {e} (maybe it was deleted very quickly)")
                        
                        continue
                
                if self.cache.get(resolved_path) is None:
                    logger.info(f"New file found: {format_file(new_file)} ")
                    self.cache[resolved_path] = File.empty(resolved_path)
                
                if self.current_file is None:
                    self.current_file = new_file
                    
                    continue
                
                if self.current_file.path != resolved_path:
                    logger.debug(f"File changed (was {self.current_file.path}, now {resolved_path})")
                    event = process_file_change(
                        new_file=self.current_file,
                        old_file=self.cache[self.current_file.path],
                        time=time.time(),
                        env=self.env
                    )
                    
                    if event is not None:
                        await queue.put(event)
                    
                    self.cache[resolved_path] = new_file
                    
                self.current_file = new_file
        
    async def setup(self, tg: TaskGroup, emit_event: Queue[Event]):
        self.task = tg.create_task(self._task(emit_event))
    
    async def shutdown(self):
        pass
    
    async def resolve_events(self) -> AsyncGenerator[Event, None]:
        if self.current_file is not None:
            event = process_file_change(
                new_file=self.current_file,
                old_file=self.cache[self.current_file.path],
                time=time.time(),
                env=self.env
            )
            
            if event is not None:
                yield event
            
            self.cache[self.current_file.path] = self.current_file
            self.current_file = None



from os import environ
from pathlib import Path

from pydantic import BaseModel
from pydantic.deprecated.class_validators import root_validator
from pydantic_settings import BaseSettings
from pydantic_settings.main import SettingsConfigDict

from wakapi_anyide.helpers.ini import IniConfigSettingsSource


class Settings(BaseModel):
    api_key: str | None = None
    api_key_vault_cmd: str | None = None
    api_url: str = "https://api.wakatime.com/api/v1"
    hostname: str | None = None
    heartbeat_rate_limit_seconds: int = 120
    
    @root_validator(pre=True)
    def validate_api_key(cls, values):
        if key := environ.get("WAKATIME_API_KEY"):
            values["api_key"] = key
        
        if values.get('api_key') == None and values.get('api_key_vault_cmd') == None:
            raise ValueError('One of api_key or api_key_vault_cmd must be set.')
        
        return values


class WakatimeConfig(BaseSettings):
    model_config = SettingsConfigDict(
        ini_file=Path(environ.get("WAKATIME_HOME", environ.get("HOME", "~/"))) / ".wakatime.cfg",
        extra='ignore'
    )  # type: ignore
    
    settings: Settings
    
    @classmethod
    def settings_customise_sources(cls, settings_cls, *_args, **_kwargs):
        return (IniConfigSettingsSource(settings_cls), )



import asyncio
from collections.abc import AsyncIterator
from dataclasses import dataclass
from typing import Protocol
from typing import Tuple

from wakapi_anyide.models.environment import Environment


@dataclass
class Event:
    filename: str
    file_extension: str
    cursor: Tuple[int, int]
    lines_added: int
    lines_removed: int
    lines: int
    time: float


class Watcher(Protocol):
    """
    A watcher implements this protocol.
    A watcher can either choose to emit events in realtime with the emit_event queue, or use resolve_events as a generator.
    """
    
    def __init__(self, env: Environment):
        raise NotImplementedError
    
    async def setup(self, tg: asyncio.TaskGroup, emit_event: asyncio.Queue[Event]):
        """
        Create long-running tasks with the tg TaskGroup.
        """
        return
    
    async def shutdown(self):
        """
        Teardown any long-running tasks here.
        """
        return
    
    def resolve_events(self) -> AsyncIterator[Event] | None:
        return



import asyncio
from wakapi_anyide._watch import Watch

async def main():
    notify = Watch()
    notify.add_watch("./")
    
    async for evs in notify:
        for ev in evs:
            print(ev)
        
if __name__ == "__main__":
    asyncio.run(main())